package at.ac.uibk.gitsearch.service;

import static org.assertj.core.api.Assertions.assertThat;

import at.ac.uibk.gitsearch.IntegrationTest;
import at.ac.uibk.gitsearch.domain.ChatMessage;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.repository.ReviewRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.dto.ChatMessageDTO;
import java.util.List;
import java.util.Optional;
import javax.transaction.Transactional;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithMockUser;

@IntegrationTest
public class ChatMessageServiceIT {

    @Autowired
    private ChatMessageService chatMessageService;

    @Autowired
    private UserService userService;

    @Autowired
    private ReviewRepository reviewRepository;

    private Review review;

    @BeforeEach
    public void setUp() {
        review = new Review();
        review.setResource("TestResource");
        reviewRepository.save(review); // You might need to mock this as well if the actual DB call isn't required.
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void testSaveAndGetChatMessage() {
        ChatMessageDTO dto = new ChatMessageDTO();
        dto.setMessageText("Test message");
        dto.setReviewID(review.getId());

        ChatMessage savedMessage = chatMessageService.saveChatMessage(dto);
        Optional<ChatMessage> retrievedMessage = chatMessageService.getChatMessageById(savedMessage.getId());

        assertThat(retrievedMessage).isPresent();
        assertThat(retrievedMessage.get().getMessageText()).isEqualTo("Test message");
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void testDeleteChatMessage() {
        ChatMessageDTO dto = new ChatMessageDTO();
        dto.setMessageText("Test message to delete");
        dto.setReviewID(review.getId());

        ChatMessage savedMessage = chatMessageService.saveChatMessage(dto);
        chatMessageService.deleteChatMessage(savedMessage.getId());
        Optional<ChatMessage> retrievedMessage = chatMessageService.getChatMessageById(savedMessage.getId());

        assertThat(retrievedMessage).isEmpty();
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void testFindByReview() {
        ChatMessageDTO dto = new ChatMessageDTO();
        dto.setMessageText("Test message for review");
        dto.setReviewID(review.getId());

        chatMessageService.saveChatMessage(dto);
        List<ChatMessage> messages = chatMessageService.findByReview(review);

        assertThat(messages).hasSize(1);
        assertThat(messages.get(0).getMessageText()).isEqualTo("Test message for review");
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void testHasAccessToReview() {
        User user = userService.getUserWithAuthorities().get();
        review.setRequestedBy(user.getId());
        reviewRepository.save(review);

        boolean hasAccess = chatMessageService.hasAccessToReview(user, review);
        assertThat(hasAccess).isTrue();
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void testToDTO() {
        ChatMessage chatMessage = new ChatMessage();
        chatMessage.setReview(review);
        chatMessage.setMessageText("Test message DTO");
        chatMessage.setSender(userService.getUserWithAuthorities().get());

        ChatMessageDTO dto = chatMessageService.toDTO(chatMessage);

        assertThat(dto.getMessageText()).isEqualTo("Test message DTO");
        assertThat(dto.getReviewID()).isEqualTo(review.getId());
        assertThat(dto.getSenderName()).isEqualTo("You");
    }
}
