package at.ac.uibk.gitsearch.service;

import static org.assertj.core.api.Assertions.assertThat;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.service.ZipRepackagingServiceTest.MockedSingletonTreeNodeForFiltering;
import io.netty.util.internal.StringUtil;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import org.junit.Assert;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithMockUser;

@SpringBootTest(classes = GitsearchApp.class)
@WithMockUser(value = "test", authorities = "sharing")
public class ZipRepackagingServiceIT {

    private static final String TEST_ZIP_LOCATION = "./testData/junit-quality-tests-exercise-master.zip";
    private static final String TEST_PYTHON_ZIP_LOCATION = "./testData/pythonprogrammierkurs-main.zip";

    @Autowired
    private ZipRepackagingService zipRepackagingService;

    @SuppressWarnings("unused")
    private static final Logger log = LoggerFactory.getLogger(ZipRepackagingServiceIT.class);

    @Test
    @Timeout(value = 10 * 60 * 1000L)
    void testSimpleRepackage() throws IOException {
        final List<String> zippedEntries = getZippedEntries(new ZipInputStream(this.getClass().getResourceAsStream(TEST_ZIP_LOCATION)))
            .stream()
            .map(s -> "/" + s.substring("junit-quality-tests-exercise-master".length() + 1))
            .filter(s -> !StringUtil.isNullOrEmpty(s))
            .collect(Collectors.toList());

        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);
        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = zipRepackagingService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            Optional.empty(),/* no filtering */
            ""
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        assertThat(getZippedEntries(testRepackagedZip)).containsAll(zippedEntries);

        testRepackagedZip.close();
    }

    @Test
    void testFilteredRepackage() throws IOException {
        final MockedSingletonTreeNodeForFiltering filterOut = new MockedSingletonTreeNodeForFiltering("solution", "tests");

        // first check whether this is good test data
        final ZipInputStream originalZip = new ZipInputStream(this.getClass().getResourceAsStream(TEST_ZIP_LOCATION));

        boolean found = false;
        for (ZipEntry entry = originalZip.getNextEntry(); entry != null; entry = originalZip.getNextEntry()) {
            found = found || isPathPrefix(filterOut.getFilterStrings(), "junit-quality-tests-exercise-master/", entry);
            if (found) break;
        }
        Assert.assertTrue("test zip should contain entries to filter away?", found);

        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = zipRepackagingService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            Optional.of(filterOut),
            ""
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        for (ZipEntry entry = testRepackagedZip.getNextEntry(); entry != null; entry = testRepackagedZip.getNextEntry()) {
            boolean foundPrefix = isPathPrefix(filterOut.getFilterStrings(), "", entry);
            Assert.assertFalse("original folder starts with excepted directory?", foundPrefix);
        }

        testRepackagedZip.close();
    }

    /**
     * tests for the path prefix in filterOut
     * @param filterOut the list of path prefixes to filter out
     * @param basePrefix a basePrefix to ignore
     * @param entry the zip entry
     * @return true if one of the path prefixes matches
     */
    private boolean isPathPrefix(final String[] filterOut, String basePrefix, ZipEntry entry) {
        boolean found = false;
        String fileName = entry.getName();
        if (fileName.startsWith(basePrefix)) fileName = fileName.substring(basePrefix.length());
        for (String filter : filterOut) {
            found = found || fileName.startsWith(filter); // this is slighlty lazy, because we should check for real path
            if (found) break;
        }
        return found;
    }

    /**
     * tests for subpackages
     * @throws IOException
     */
    @Test
    void testSubRepackage() throws IOException {
        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        final InputStream testRepackagedZipIS = zipRepackagingService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_ZIP_LOCATION,
            Optional.empty(),/* no filtering */
            "exercise"
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        for (ZipEntry entry = testRepackagedZip.getNextEntry(); entry != null; entry = testRepackagedZip.getNextEntry()) {
            boolean foundPrefix = isPathPrefix(new String[] { "solution", "tests", "test" }, "", entry);
            Assert.assertFalse("original folder starts with excepted directory?", foundPrefix);
        }

        testRepackagedZip.close();
    }

    private MockedSingletonTreeNodeForFiltering constructTreeNodeForFiltering() {
        MockedSingletonTreeNodeForFiltering root = new MockedSingletonTreeNodeForFiltering(
            "metadata.yml",
            true,
            "/infrastructure",
            "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb",
            "/SS21_Assignment1/SimplePrograms_Variables/SimplePrograms_Variables_mit_musterloesung.ipynb",
            "/SS21_Assignment1/SimplePrograms_AddedValueTax/SimplePrograms_AddedValueTax_mit_musterloesung.ipynb",
            "/SS21_Assignment1/SimplePrograms_Calculations/SimplePrograms_Calculations_mit_musterloesung.ipynb",
            "/SS21_Assignment1/SimplePrograms_OfferCreator/SimplePrograms_OfferCreator_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_LargestNumber/Conditions_LargestNumber_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_NegativePositiveZero/Conditions_NegativePositiveZero_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_DiscountCalculator/Conditions_DiscountCalculator_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_Quarter/Conditions_Quarter_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_Calculator/Conditions_Calculator_mit_musterloesung.ipynb",
            "/SS23_Assignment2/Conditions_WageTax/Conditions_WageTax_mit_musterloesung.ipynb"
        );
        MockedSingletonTreeNodeForFiltering assignment1 = new MockedSingletonTreeNodeForFiltering(
            "SS21_Assignment1/metadata.yml",
            true,
            "/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb",
            "/SimplePrograms_Variables/SimplePrograms_Variables_mit_musterloesung.ipynb",
            "/SimplePrograms_AddedValueTax/SimplePrograms_AddedValueTax_mit_musterloesung.ipynb",
            "/SimplePrograms_Calculations/SimplePrograms_Calculations_mit_musterloesung.ipynb",
            "/SimplePrograms_OfferCreator/SimplePrograms_OfferCreator_mit_musterloesung.ipynb"
        );
        MockedSingletonTreeNodeForFiltering helloWorld = new MockedSingletonTreeNodeForFiltering(
            "SS21_Assignment1/SimplePrograms_HelloWorld/metadata.yml",
            true,
            "/SimplePrograms_HelloWorld_mit_musterloesung.ipynb",
            "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb"
        );
        assignment1.addChild(helloWorld);
        root.addChild(assignment1);

        return root;
    }

    private static List<String> expectedEntriesAll = List.of(
        "/README.md",
        "/SS21_Assignment1/",
        "/SS21_Assignment1/SimplePrograms_AddedValueTax/",
        "/SS21_Assignment1/SimplePrograms_AddedValueTax/SimplePrograms_AddedValueTax.ipynb",
        "/SS21_Assignment1/SimplePrograms_AddedValueTax/metadata.yaml",
        "/SS21_Assignment1/SimplePrograms_Calculations/",
        "/SS21_Assignment1/SimplePrograms_Calculations/SimplePrograms_Calculations.ipynb",
        "/SS21_Assignment1/SimplePrograms_Calculations/metadata.yaml",
        "/SS21_Assignment1/SimplePrograms_HelloWorld/",
        "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld.ipynb",
        "/SS21_Assignment1/SimplePrograms_HelloWorld/metadata.yaml",
        "/SS21_Assignment1/SimplePrograms_OfferCreator/",
        "/SS21_Assignment1/SimplePrograms_OfferCreator/SimplePrograms_OfferCreator.ipynb",
        "/SS21_Assignment1/SimplePrograms_OfferCreator/metadata.yaml",
        "/SS21_Assignment1/SimplePrograms_Variables/",
        "/SS21_Assignment1/SimplePrograms_Variables/SimplePrograms_Variables.ipynb",
        "/SS21_Assignment1/SimplePrograms_Variables/metadata.yaml",
        "/SS21_Assignment1/metadata.yaml",
        "/SS23_Assignment2/",
        "/SS23_Assignment2/Conditions_Calculator/",
        "/SS23_Assignment2/Conditions_Calculator/Conditions_Calculator.ipynb",
        "/SS23_Assignment2/Conditions_Calculator/metadata.yaml",
        "/SS23_Assignment2/Conditions_DiscountCalculator/",
        "/SS23_Assignment2/Conditions_DiscountCalculator/Conditions_DiscountCalculator.ipynb",
        "/SS23_Assignment2/Conditions_DiscountCalculator/metadata.yaml",
        "/SS23_Assignment2/Conditions_LargestNumber/",
        "/SS23_Assignment2/Conditions_LargestNumber/Conditions_LargestNumber.ipynb",
        "/SS23_Assignment2/Conditions_LargestNumber/metadata.yaml",
        "/SS23_Assignment2/Conditions_NegativePositiveZero/",
        "/SS23_Assignment2/Conditions_NegativePositiveZero/Conditions_NegativePositiveZero.ipynb",
        "/SS23_Assignment2/Conditions_NegativePositiveZero/metadata.yaml",
        "/SS23_Assignment2/Conditions_Quarter/",
        "/SS23_Assignment2/Conditions_Quarter/Conditions_Quarter.ipynb",
        "/SS23_Assignment2/Conditions_Quarter/metadata.yaml",
        "/SS23_Assignment2/Conditions_WageTax/",
        "/SS23_Assignment2/Conditions_WageTax/Conditions_WageTax.ipynb",
        "/SS23_Assignment2/Conditions_WageTax/metadata.yaml",
        "/SS23_Assignment2/metadata.yaml",
        "/jku.png",
        "/jku.svg",
        "/metadata.yaml",
        "/requirements.txt"
    );

    private static List<String> expectedEntriesAssignment1 = List.of(
        "/SimplePrograms_AddedValueTax/",
        "/SimplePrograms_AddedValueTax/SimplePrograms_AddedValueTax.ipynb",
        "/SimplePrograms_AddedValueTax/metadata.yaml",
        "/SimplePrograms_Calculations/",
        "/SimplePrograms_Calculations/SimplePrograms_Calculations.ipynb",
        "/SimplePrograms_Calculations/metadata.yaml",
        "/SimplePrograms_HelloWorld/",
        "/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld.ipynb",
        "/SimplePrograms_HelloWorld/metadata.yaml",
        "/SimplePrograms_OfferCreator/",
        "/SimplePrograms_OfferCreator/SimplePrograms_OfferCreator.ipynb",
        "/SimplePrograms_OfferCreator/metadata.yaml",
        "/SimplePrograms_Variables/",
        "/SimplePrograms_Variables/SimplePrograms_Variables.ipynb",
        "/SimplePrograms_Variables/metadata.yaml",
        "/metadata.yaml"
    );
    private static List<String> expectedEntriesHelloWorld = List.of("/SimplePrograms_HelloWorld.ipynb", "/metadata.yaml");

    static Stream<Arguments> checkExplicitMethodSourceArgs() {
        return Stream.of(
            Arguments.of("", expectedEntriesAll),
            Arguments.of("SS21_Assignment1", expectedEntriesAssignment1),
            Arguments.of("SS21_Assignment1/SimplePrograms_HelloWorld", expectedEntriesHelloWorld)
        );
    }

    @Test
    void testGetApplicableFilters() {
        final MockedSingletonTreeNodeForFiltering filterOut = constructTreeNodeForFiltering();
        final List<String> applicableFiltersAll = ZipRepackagingService.getApplicableFilters("", filterOut);
        final List<String> applicableFiltersAssignment = ZipRepackagingService.getApplicableFilters("SS21_Assignment1", filterOut);
        final List<String> applicableFiltersHelloWorld = ZipRepackagingService.getApplicableFilters(
            "SS21_Assignment1/SimplePrograms_HelloWorld",
            filterOut
        );
        assertThat(applicableFiltersAll)
            .containsAll(
                List.of("/infrastructure", "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb")
            );
        assertThat(applicableFiltersAssignment)
            .containsAll(
                List.of(
                    "/infrastructure",
                    "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb",
                    "/SS21_Assignment1/SimplePrograms_Variables/SimplePrograms_Variables_mit_musterloesung.ipynb"
                )
            );
        assertThat(applicableFiltersHelloWorld)
            .containsAll(
                List.of(
                    "/infrastructure",
                    "/SS21_Assignment1/SimplePrograms_HelloWorld/SimplePrograms_HelloWorld_mit_musterloesung.ipynb",
                    "/SS21_Assignment1/SimplePrograms_Variables/SimplePrograms_Variables_mit_musterloesung.ipynb"
                )
            );
    }

    @ParameterizedTest
    @Timeout(value = 10 * 60 * 1000L)
    @MethodSource("checkExplicitMethodSourceArgs")
    void testComplexFilteredRepackage(String rootPath, List<String> expected) throws IOException {
        final MockedSingletonTreeNodeForFiltering filterOut = constructTreeNodeForFiltering();

        final InputStream resourceStream = this.getClass().getResourceAsStream(TEST_PYTHON_ZIP_LOCATION);

        Assert.assertNotNull("cannot find resource at " + TEST_PYTHON_ZIP_LOCATION, resourceStream);
        final ZipInputStream testZip = new ZipInputStream(resourceStream);

        // first complete return (without
        final InputStream testRepackagedZipIS = zipRepackagingService.rePackageGitLabProjectZip(
            testZip,
            "from " + TEST_PYTHON_ZIP_LOCATION,
            Optional.of(filterOut),
            rootPath
        );

        final ZipInputStream testRepackagedZip = new ZipInputStream(testRepackagedZipIS);

        assertThat(getZippedEntries(testRepackagedZip)).containsAll(expected);

        testRepackagedZip.close();
    }

    public List<String> getZippedEntries(ZipInputStream testZip) throws IOException {
        List<String> result = new ArrayList<>();

        ZipEntry nextEntry = testZip.getNextEntry();

        while (nextEntry != null) {
            result.add(nextEntry.getName());
            nextEntry = testZip.getNextEntry();
        }
        testZip.close(); // it does not make sense, to leave it open
        return result;
    }
}
